#resource "\\Indicators\\Spy.ex5" // https://www.mql5.com/ru/forum/170952/page68#comment_6406566

#include <fxsaber\ThirdPartyTicks\File.mqh> // https://www.mql5.com/ru/code/20225
#include <fxsaber\Expert.mqh> // https://www.mql5.com/ru/code/19003
#include <fxsaber\MultiTester\MTTester.mqh> // https://www.mql5.com/ru/code/26132

#include <fxsaber\SingleTesterCache\SingleTesterCache.mqh> // https://www.mql5.com/ru/code/27611
#include <MT4Orders.mqh> // https://www.mql5.com/ru/code/16006

class VALIDATEPORTFOLIO
{
private:
  bool FirstRun;
  bool RunFlag;
  const string FolderName;
  const bool IsTester;
  const string PortfolioFolder;

  string PortfolioString;

  SINGLETESTERCACHE Portfolio[];
  int PosDeal[];

  string Symbols[];

  static int GetLastPos( const string &Str, const short Char )
  {
    int Pos = ::StringLen(Str) - 1;

    while ((Pos >= 0) && (Str[Pos] != Char))
      Pos--;

    return(Pos);
  }

  static string GetName( const string &Path )
  {
    const int BeginPos = VALIDATEPORTFOLIO::GetLastPos(Path, '\\') + 1;
    const int LastPos = VALIDATEPORTFOLIO::GetLastPos(Path, '.');

    return((LastPos > BeginPos) ? ::StringSubstr(Path, BeginPos, LastPos - BeginPos) : NULL);
  }

  int SetFiles( const string &FileNames[] ) const
  {
    string FileNamesDelete[];

    for (int i = FILE::GetFileNames(FileNamesDelete, this.PortfolioFolder + "*.tst", FILE_COMMON) - 1; i >= 0; i--)
      ::FileDelete(FileNamesDelete[i], FILE_COMMON);

    const int Size = ::ArraySize(FileNames);

    for (int i = 0; i < Size; i++)
      ::FileCopy(FileNames[i], 0, this.PortfolioFolder + VALIDATEPORTFOLIO::GetName(FileNames[i]) + ".tst", FILE_COMMON);

    return(Size);
  }

  int GetPortfolio( const string &FileNames[], const int Common_Flag = 0 )
  {
    int Amount = 0;
    const int Size = ::ArrayResize(this.Portfolio, ::ArraySize(FileNames));

    for (int i = 0; i < Size; i++)
      if (this.Portfolio[Amount].Load(FileNames[i], Common_Flag))
        Amount++;

    return(::ArrayResize(Portfolio, Amount));
  }

  int GetPortfolioSymbols( void )
  {
    int Size = 0;

    ::ArrayFree(this.Symbols);

    for (int i = ::ArraySize(this.Portfolio) - 1; i >= 0; i--)
    {
      string Symbols2[];

      for (int j = this.Portfolio[i].GetTradeSymbols(Symbols2) - 1; j >= 0; j--)
      {
        const string Symb = Symbols2[j];

        int k = 0;

        for (; (k < Size) && (Symbols[k] != Symb); k++)
          ;

        if (k == Size)
        {
          Size = ::ArrayResize(Symbols, Size + 1);

          Symbols[Size - 1] = Symb;
        }
      }
    }

    return(Size);
  }

  string PortfolioToString( const string &FileNames[], const int MaxAmount = INT_MAX ) const
  {
    const int Size = ::ArraySize(this.Portfolio);
    const int Amount = ::ArraySize(this.Symbols);

    string Str = "TesterPortfolio:";

    for (int i = 0; i < Size; i++)
      Str += "\n" + (string)(i + 1) + ": "+ FileNames[i] + "\n" + ((Size < MaxAmount) ? this.Portfolio[i].ToString() + "\n" : NULL);

    Str += "\nSymbols:";

    for (int i = 0; i < Amount; i++)
      Str += "\n" + (string)(i + 1) + ": "+ this.Symbols[i];

    return(Str);
  }

  void GetData( double &Deposit, datetime &FromDate, datetime &ToDate ) const
  {
    Deposit = 1e5;
    FromDate = INT_MAX;
    ToDate = 0;

    for (int i = ::ArraySize(this.Portfolio) - 1; i >= 0; i--)
    {
      if (this.Portfolio[i].Header.date_from < FromDate)
      {
        Deposit = this.Portfolio[i].Header.trade_deposit;

        FromDate = (datetime)this.Portfolio[i].Deals[0].GetProperty(DEAL_TIME);
      }

      if (this.Portfolio[i].Header.date_to > ToDate)
        ToDate = this.Portfolio[i].Header.date_to;
    }

    return;
  }

  bool RunMe( void )
  {
    if (this.FirstRun)
    {
      string FileNames[];

      this.RunFlag = !FILE::GetFileNames(FileNames, this.FolderName + "*.tst") ||
                     !this.GetPortfolio(FileNames) ||
                     !this.GetPortfolioSymbols() ||
                     !(::MessageBox(this.PortfolioToString(FileNames, 5), __FILE__,  MB_OKCANCEL) == IDOK) ||
                     !this.SetFiles(FileNames);

      this.FirstRun = false;
    }

    if (!this.RunFlag)
    {
      MqlParam Params[];
      string Names[];

      bool Res = EXPERT::Parameters(0, Params, Names, false);

      if (Res)
      {
        double Deposit;
        datetime FromDate;
        datetime ToDate;

        this.GetData(Deposit, FromDate, ToDate);

        string Str = "[Tester]" +
                     "\nExpert=" + ::StringSubstr(Params[0].string_value, ::StringLen("Experts\\")) +
                     "\nSymbol=" + this.Symbols[0] +
                     "\nOptimization=0" +
                     "\nModel=4" +
                     "\nFromDate=" + ::TimeToString(FromDate, TIME_DATE) +
                     "\nToDate=" + ::TimeToString(ToDate, TIME_DATE) +
                     "\nDeposit=" + ::DoubleToString(Deposit, 2) +
                     "\n[TesterInputs]";

        const int Size = ::ArraySize(Names);

        for (int i = 0; i < Size; i++)
          Str += "\n" + Names[i] + "=" + Params[i + 1].string_value;

        Res = MTTESTER::SetSettings2(Str) && MTTESTER::ClickStart();
      }

      this.RunFlag = Res;
    }

    return(this.RunFlag);
  }

  static bool OrderScan( const int Type, const string Symb, const MAGIC_TYPE Magic, const double Lot = 0 )
  {
    double LotsMax = Lot;
    TICKET_TYPE TicketMax = 0;

    for (int i = OrdersTotal() - 1; i >= 0; i--)
      if (OrderSelect(i, SELECT_BY_POS) && (OrderType() == Type) && (OrderMagicNumber() == Magic) && (OrderSymbol() == Symb))
      {
        if (OrderLots() == Lot)
          return(true);
        else if (OrderLots() > LotsMax)
        {
          LotsMax = OrderLots();

          TicketMax = OrderTicket();
        }
      }

    return(TicketMax && OrderSelect(TicketMax, SELECT_BY_TICKET));
  }

  static bool CloseOrder( const string Symb, const int Type, double Lots, const double Price, const MAGIC_TYPE Magic, const string comment )
  {
    bool Res = Lots;

    while (Res && Lots && (Res = VALIDATEPORTFOLIO::OrderScan(Type, Symb, Magic, Lots)))
    {
      Res = (OrderClosePrice() == Price) && OrderClose(OrderTicket(), Lots, Price, 0, clrNONE, comment);

      Lots = ::NormalizeDouble(Lots - OrderLots(), 8);
    }

    return(Res);
  }

  #define MACROS(A) \
  {                 \
    if (!(A))       \
    {               \
      Print(#A);    \
                    \
      TesterStop(); \
    }               \
  }

  void PortfolioTrade()
  {
    static const bool IsNetting = ((ENUM_ACCOUNT_MARGIN_MODE)::AccountInfoInteger(ACCOUNT_MARGIN_MODE) != ACCOUNT_MARGIN_MODE_RETAIL_HEDGING);

    if (this.IsTester)
    {
      const int Size = ::ArraySize(this.Portfolio);
      const datetime time = ::TimeTradeServer();

      for (int i = 0; i < Size; i++)
      {
        bool Flag = true;

        while (Flag &&
               (this.PosDeal[i] < ::ArraySize(this.Portfolio[i].Deals)) &&
               (this.Portfolio[i].Deals[this.PosDeal[i]].GetProperty(DEAL_TIME) <= time))
        {
          const TradeDeal Deal = this.Portfolio[i].Deals[this.PosDeal[i]];
          const datetime DealTime = (datetime)Deal.GetProperty(DEAL_TIME);

          if (Deal.GetProperty(DEAL_TIME) == time)
          {
            const int Type = (int)Deal.GetProperty(DEAL_TYPE);

            if (Type == OP_BALANCE)
            {
              const double Profit = Deal.GetProperty(DEAL_PROFIT);

              if (Profit > 0)
                MACROS(::TesterDeposit(Profit))
              else if (Profit < 0)
                MACROS(::TesterWithdrawal(-Profit))
            }
            else
            {
              const string Symb = Deal.GetProperty(DEAL_SYMBOL);
              const double Price = ::SymbolInfoDouble(Symb, Type ? SYMBOL_BID : SYMBOL_ASK);

              if (Flag = !::NormalizeDouble(Deal.GetProperty(DEAL_PRICE) - Price, 8))
  //            if (Flag = (Deal.GetProperty(DEAL_PRICE) == Price))
              {
                if (IsNetting || (Deal.GetProperty(DEAL_ENTRY) == DEAL_ENTRY_IN))
                  MACROS(OrderSend(Symb, Type, Deal.GetProperty(DEAL_VOLUME), Price, 0, 0, 0, Deal.GetProperty(DEAL_COMMENT), i) != -1)
                else
                  MACROS(CloseOrder(Symb, 1 - Type, Deal.GetProperty(DEAL_VOLUME), Price, i, Deal.GetProperty(DEAL_COMMENT)))
              }
            }

            if (Flag)
              this.PosDeal[i]++;
          }
          else
            this.PosDeal[i]++;
        }
      }
    }
  }

public:
  VALIDATEPORTFOLIO( const string sFolderName ) : FolderName(sFolderName), FirstRun(true), RunFlag(false),
                                                  IsTester(::MQLInfoInteger(MQL_TESTER)),
                                                  PortfolioFolder(__FILE__ + "\\"), PortfolioString(NULL)
  {
    if (this.IsTester)
    {
      string FileNames[];

      if (!(this.RunFlag = !FILE::GetFileNames(FileNames, this.PortfolioFolder + "*.tst", FILE_COMMON) ||
                           !::ArrayResize(this.PosDeal, this.GetPortfolio(FileNames, FILE_COMMON)) ||
                           !this.GetPortfolioSymbols()))

      {
        for (int i = ::ArraySize(this.Symbols) - 1; i >= 0; i--)
          if (this.Symbols[i] != _Symbol)
            ::iCustom(Symbols[i], PERIOD_W1, "::Indicators\\Spy.ex5", ::ChartID(), i);

        this.PortfolioString = this.PortfolioToString(FileNames);

        ::ArrayInitialize(this.PosDeal, 1);

        this.PortfolioTrade();
      }
    }
  }

  ~VALIDATEPORTFOLIO( void )
  {
    if (this.IsTester && !this.RunFlag)
      ::Print(this.PortfolioString);
  }

  bool IsFinish( void ) const
  {
    return(this.RunFlag && (IsTester || MTTESTER::IsReady()));
  }

  void OnTimer( void )
  {
    if (!this.IsTester)
      this.RunMe();
    else if (!this.RunFlag)
      this.PortfolioTrade();
  }

  void OnTick()
  {
    if (this.IsTester && !this.RunFlag)
      this.PortfolioTrade();

    return;
  }

  void OnChartEvent( const int id, const long&, const double&, const string& )
  {
    if (this.IsTester && !this.RunFlag && (id == CHARTEVENT_CUSTOM))
      this.PortfolioTrade();

    return;
  }
};